@@ -63,6 +63,9 @@ extern int statvfs (const char * file, /*@out@*/ struct statvfs * buf)
 /*@unchecked@*/
 int _rpmts_debug = 0;
 
+/*@unchecked@*/
+int _rpmts_stats = 0;
+
 char * hGetNEVR(Header h, const char ** np)
 {
     const char * n, * v, * r;
@@ -179,15 +182,113 @@ int rpmtsVerifyDB(rpmts ts)
     return rpmdbVerify(ts->rootDir);
 }
 
+static int isArch(const char * arch)
+	/*@*/
+{
+    const char ** av;
+    static const char *arches[] = {
+	"i386", "i486", "i586", "i686", "athlon", "x86_64",
+	"alpha", "alphaev5", "alphaev56", "alphapca56", "alphaev6", "alphaev67",
+	"sparc", "sun4", "sun4m", "sun4c", "sun4d", "sparcv9",
+	"sparc64", "sun4u",
+	"mips", "mipsel", "IP",
+	"ppc", "ppciseries", "ppcpseries",
+	"ppc64", "ppc64iseries", "ppc64pseries",
+	"m68k",
+	"rs6000",
+	"ia64",
+	"armv3l", "armv4b", "armv4l",
+	"s390", "i370", "s390x",
+	"sh", "xtensa",
+	"noarch",
+	NULL,
+    };
+
+    for (av = arches; *av != NULL; av++) {
+	if (!strcmp(arch, *av))
+	    return 1;
+    }
+    return 0;
+}
+
 rpmdbMatchIterator rpmtsInitIterator(const rpmts ts, rpmTag rpmtag,
 			const void * keyp, size_t keylen)
 {
     rpmdbMatchIterator mi;
+    const char * arch = NULL;
+    int xx;
+
     if (ts->rdb == NULL && rpmtsOpenDB(ts, ts->dbmode))
 	return NULL;
+
+    /* Parse out "N(EVR).A" tokens from a label key. */
+/*@-branchstate@*/
+    if (rpmtag == RPMDBI_LABEL && keyp != NULL) {
+	const char * s = keyp;
+	const char *se;
+	size_t slen = strlen(s);
+	char *t = alloca(slen+1);
+	int level = 0;
+	int c;
+
+	keyp = t;
+	while ((c = *s++) != '\0') {
+	    switch (c) {
+	    default:
+		*t++ = c;
+		break;
+	    case '(':
+		/* XXX Fail if nested parens. */
+		if (level++ != 0) {
+		    rpmError(RPMERR_QFMT, _("extra '(' in package label: %s\n"), keyp);
+		    return NULL;
+		}
+		/* Parse explicit epoch. */
+		for (se = s; *se && xisdigit(*se); se++)
+		    ;
+		if (*se == ':') {
+		    /* XXX skip explicit epoch's (for now) */
+		    *t++ = '-';
+		    s = se + 1;
+		} else {
+		    /* No Epoch: found. Convert '(' to '-' and chug. */
+		    *t++ = '-';
+		}
+		break;
+	    case ')':
+		/* XXX Fail if nested parens. */
+		if (--level != 0) {
+		    rpmError(RPMERR_QFMT, _("missing '(' in package label: %s\n"), keyp);
+		    return NULL;
+		}
+		/* Don't copy trailing ')' */
+		break;
+	    }
+	}
+	if (level) {
+	    rpmError(RPMERR_QFMT, _("missing ')' in package label: %s\n"), keyp);
+	    return NULL;
+	}
+	*t = '\0';
+	t = (char *) keyp;
+	t = strrchr(t, '.');
+	/* Is this a valid ".arch" suffix? */
+	if (t != NULL && isArch(t+1)) {
+	   *t++ = '\0';
+	   arch = t;
+	}
+    }
+/*@=branchstate@*/
+
     mi = rpmdbInitIterator(ts->rdb, rpmtag, keyp, keylen);
+
+    /* Verify header signature/digest during retrieve (if not disabled). */
     if (mi && !(ts->vsflags & RPMVSF_NOHDRCHK))
 	(void) rpmdbSetHdrChk(mi, ts, headerCheck);
+
+    /* Select specified arch only. */
+    if (arch != NULL)
+	xx = rpmdbSetIteratorRE(mi, RPMTAG_ARCH, RPMMIRE_DEFAULT, arch);
     return mi;
 }
 
@@ -627,6 +728,43 @@ void rpmtsEmpty(rpmts ts)
 /*@=nullstate@*/
 }
 
+static void rpmtsPrintStat(const char * name, /*@null@*/ struct rpmop_s * op)
+	/*@globals fileSystem @*/
+	/*@modifies fileSystem @*/
+{
+    static unsigned int scale = (1000 * 1000);
+    if (op != NULL && op->count > 0)
+	fprintf(stderr, "   %s %6d %6lu.%06lu MB %6lu.%06lu secs\n",
+		name, op->count,
+		(unsigned long)op->bytes/scale, (unsigned long)op->bytes%scale,
+		op->usecs/scale, op->usecs%scale);
+}
+
+static void rpmtsPrintStats(rpmts ts)
+	/*@globals fileSystem, internalState @*/
+	/*@modifies fileSystem, internalState @*/
+{
+    (void) rpmswExit(rpmtsOp(ts, RPMTS_OP_TOTAL), 0);
+
+    rpmtsPrintStat("total:       ", rpmtsOp(ts, RPMTS_OP_TOTAL));
+    rpmtsPrintStat("check:       ", rpmtsOp(ts, RPMTS_OP_CHECK));
+    rpmtsPrintStat("order:       ", rpmtsOp(ts, RPMTS_OP_ORDER));
+    rpmtsPrintStat("fingerprint: ", rpmtsOp(ts, RPMTS_OP_FINGERPRINT));
+    rpmtsPrintStat("repackage:   ", rpmtsOp(ts, RPMTS_OP_REPACKAGE));
+    rpmtsPrintStat("install:     ", rpmtsOp(ts, RPMTS_OP_INSTALL));
+    rpmtsPrintStat("erase:       ", rpmtsOp(ts, RPMTS_OP_ERASE));
+    rpmtsPrintStat("scriptlets:  ", rpmtsOp(ts, RPMTS_OP_SCRIPTLETS));
+    rpmtsPrintStat("compress:    ", rpmtsOp(ts, RPMTS_OP_COMPRESS));
+    rpmtsPrintStat("uncompress:  ", rpmtsOp(ts, RPMTS_OP_UNCOMPRESS));
+    rpmtsPrintStat("digest:      ", rpmtsOp(ts, RPMTS_OP_DIGEST));
+    rpmtsPrintStat("signature:   ", rpmtsOp(ts, RPMTS_OP_SIGNATURE));
+    rpmtsPrintStat("dbadd:       ", rpmtsOp(ts, RPMTS_OP_DBADD));
+    rpmtsPrintStat("dbremove:    ", rpmtsOp(ts, RPMTS_OP_DBREMOVE));
+    rpmtsPrintStat("dbget:       ", rpmtsOp(ts, RPMTS_OP_DBGET));
+    rpmtsPrintStat("dbput:       ", rpmtsOp(ts, RPMTS_OP_DBPUT));
+    rpmtsPrintStat("dbdel:       ", rpmtsOp(ts, RPMTS_OP_DBDEL));
+}
+
 rpmts rpmtsFree(rpmts ts)
 {
     if (ts == NULL)
@@ -667,6 +805,9 @@ rpmts rpmtsFree(rpmts ts)
     ts->pkpktlen = 0;
     memset(ts->pksignid, 0, sizeof(ts->pksignid));
 
+    if (_rpmts_stats)
+	rpmtsPrintStats(ts);
+
     (void) rpmtsUnlink(ts, "tsCreate");
 
     /*@-refcounttrans -usereleased @*/
@@ -1177,6 +1318,17 @@ uint_32 rpmtsSetColor(rpmts ts, uint_32 color)
     return ocolor;
 }
 
+rpmop rpmtsOp(rpmts ts, rpmtsOpX opx)
+{
+    rpmop op = NULL;
+
+    if (ts != NULL && opx >= 0 && opx < RPMTS_OP_MAX)
+	op = ts->ops + opx;
+/*@-usereleased -compdef @*/
+    return op;
+/*@=usereleased =compdef @*/
+}
+
 int rpmtsSetNotifyCallback(rpmts ts,
 		rpmCallbackFunction notify, rpmCallbackData notifyData)
 {
@@ -1222,6 +1374,8 @@ rpmts rpmtsCreate(void)
     rpmts ts;
 
     ts = xcalloc(1, sizeof(*ts));
+    memset(&ts->ops, 0, sizeof(ts->ops));
+    (void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_TOTAL), -1);
     ts->goal = TSM_UNKNOWN;
     ts->filesystemCount = 0;
     ts->filesystems = NULL;
