@@ -28,7 +28,6 @@ struct fprintCacheEntry_s {
     const char * dirName;		/*!< path to existing directory */
     dev_t dev;				/*!< stat(2) device number */
     ino_t ino;				/*!< stat(2) inode number */
-    int isFake;				/*!< (currently unused) */
 };
 
 /**
@@ -50,10 +49,6 @@ struct fingerPrint_s {
 /*@dependent@*/ const char * baseName;	/*!< file base name */
 };
 
-/* only if !scarceMemory */
-/** */
-#define fpFree(a) free((void *)(a).baseName)
-
 /** */
 #define	FP_ENTRY_EQUAL(a, b) (((a)->dev == (b)->dev) && ((a)->ino == (b)->ino))
 
@@ -80,8 +75,9 @@ extern "C" {
  */
 int rpmdbFindFpList(/*@null@*/ rpmdb db, fingerPrint  * fpList,
 		/*@out@*/ dbiIndexSet * matchList, int numItems)
-	/*@globals fileSystem@*/
-	/*@modifies db, *matchList, fileSystem @*/;
+	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@modifies db, *matchList, rpmGlobalMacroContext,
+		fileSystem, internalState @*/;
 
 /* Be carefull with the memory... assert(*fullName == '/' || !scareMemory) */
 
@@ -96,8 +92,10 @@ int rpmdbFindFpList(/*@null@*/ rpmdb db, fingerPrint  * fpList,
 /**
  * Destroy finger print cache.
  * @param cache		pointer to fingerprint cache
+ * @return		NULL always
  */
-void fpCacheFree(/*@only@*/ fingerPrintCache cache)
+/*@null@*/
+fingerPrintCache fpCacheFree(/*@only@*/ fingerPrintCache cache)
 	/*@modifies cache @*/;
 
 /**
@@ -123,7 +121,7 @@ unsigned int fpHashFunction(const void * key)
 
 /**
  * Compare two finger print entries.
- * exactly equivalent to FP_EQUAL macro.
+ * This routine is exactly equivalent to the FP_EQUAL macro.
  * @param key1		finger print 1
  * @param key2		finger print 2
  * @return result of comparing key1 and key2
@@ -146,17 +144,6 @@ void fpLookupList(fingerPrintCache cache, const char ** dirNames,
 		  int fileCount, fingerPrint * fpList)
 	/*@modifies cache, *fpList @*/;
 
-/**
- * Return finger prints of all file names in header.
- * @warning: scareMemory is assumed!
- * @param cache		pointer to fingerprint cache
- * @param h		package header
- * @retval fpList	pointer to array of finger prints
- */
-/*@unused@*/
-void fpLookupHeader(fingerPrintCache cache, Header h, fingerPrint * fpList)
-	/*@modifies h, cache, *fpList @*/;
-
 #ifdef __cplusplus
 }
 #endif
